(***********************************************************************)
(*                                                                     *)
(*                               Grimage                               *)
(*                                                                     *)
(*                           Graphics image                            *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2014-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ppm.ml,v 1.13 2014/03/08 12:23:47 fclement Exp $ *)

(* Manipulating images in portable format: PPM, PGM, and PBM.

- PPM: portable pixmap (pixels (picture element) map).
- PGM: portable greymap (grey scale map).
- PBM: portable bitmap (binary digit map).

*)

open Generic_image;;
open Graphics;;

type ppm = pixmap;;

(* Reading PPM images. *)

type ppm_magic_number = | P1 | P2 | P3 | P4 | P5 | P6;;

let magic_number_of_string = function
  | "P1" -> P1
  | "P2" -> P2
  | "P3" -> P3
  | "P4" -> P4
  | "P5" -> P5
  | "P6" -> P6
  | s -> invalid_arg ("Unknown magic number for PPM image: " ^ s);;

let read_ppm_magic_number ib = Scanf.bscanf ib "%2s " magic_number_of_string;;

let ppm_magic_number fname =
  let ib = Scanf.Scanning.from_file_bin fname in
  try read_ppm_magic_number ib with
  | End_of_file -> invalid_arg "magic_number: premature end of file"
;;

let string_of_magic_number = function
  | P1 -> "P1"
  | P2 -> "P2"
  | P3 -> "P3"
  | P4 -> "P4"
  | P5 -> "P5"
  | P6 -> "P6"
;;

(* Auxiliaries: skipping comments and reading numbers into strings. *)
let rec skip_comments ib =
  Scanf.bscanf ib "%0c" (function
  | '#' -> Scanf.bscanf ib "#%_[^\n]\n" (fun () -> skip_comments ib) ()
  | _ -> ())
;;

let read_int ib = Scanf.bscanf ib " %u " (fun u -> u);;
let read_max = read_int;;

let read_dims ib = Scanf.bscanf ib "%u %u ";;

let read_ppm_header ib =
  (* P6
     # CREATOR: XV Version 3.10  Rev: 12/16/94
     256 162
     255
  *)
  let mn = read_ppm_magic_number ib in
  skip_comments ib;
  read_dims ib (fun c l -> mn, l, c)
;;

(* Reading pixmaps. *)
let read_raw_pixel24 ib =
  Scanf.bscanf ib "%c%c%c"
    (fun r g b -> Graphics.rgb (int_of_char r) (int_of_char g) (int_of_char b))
;;

let read_ascii_pixel24 ib = Scanf.bscanf ib " %u %u %u " Graphics.rgb;;

let read_raw_ppm_ib ib l c =
  let img = Array.make_matrix l c 0 in
  for i = 0 to l - 1 do
    let line = img.(i) in
    for j = 0 to c - 1 do
      line.(j) <- read_raw_pixel24 ib
    done
  done;
  img
;;

let read_ascii_ppm_ib ib l c =
  let img = Array.make_matrix l c 0 in
  for i = 0 to l - 1 do
    let line = img.(i) in
    for j = 0 to c - 1 do
      line.(j) <- read_ascii_pixel24 ib
    done
  done;
  img
;;

let input_byte ib = Scanf.bscanf ib "%c" (fun c -> int_of_char c);;

(* Reading greymaps. *)
let read_raw_grey ib =
  let g = input_byte ib in
  Graphics.rgb g g g
;;

let read_ascii_grey ib =
  let r = read_int ib in
  Graphics.rgb r r r
;;

let read_raw_pgm_ib ib l c =
  let img = Array.make_matrix l c 0 in
  for i = 0 to l - 1 do
    let line = img.(i) in
    for j = 0 to c - 1 do
      line.(j) <- read_raw_grey ib
    done
  done;
  img
;;

let read_ascii_pgm_ib ib l c =
  let img = Array.make_matrix l c 0 in
  for i = 0 to l - 1 do
    let line = img.(i) in
    for j = 0 to c - 1 do
      line.(j) <- read_ascii_grey ib
    done
  done;
  img
;;

(* Reading bitmaps. *)
let read_raw_pbm_ib ib l c =
  let img = Array.make_matrix l c 0 in
  for i = 0 to l - 1 do
    let line = img.(i) in
    let rec loop j bn byte =
      if j = c then () else
      if bn = 8 then loop j 0 (input_byte ib) else
      let color =
        match byte land 0x80 with
        | 0 -> black
        | _ -> white in
      line.(j) <- color;
      let new_byte = byte lsl 1 in
      loop (j + 1) (bn + 1) new_byte in
    loop 0 0 (input_byte ib)
  done;
  img
;;

let read_ascii_bit ib =
  Scanf.bscanf ib " %c" (function
  | '0' -> black
  | _ -> white)
;;

let read_ascii_pbm_ib ib l c =
  let img = Array.make_matrix l c 0 in
  for i = 0 to l - 1 do
    let line = img.(i) in
    for j = 0 to c - 1 do
      line.(j) <- read_ascii_bit ib
    done;
  done;
  img
;;

(* Set the transparency mask: if pixel i j of mask is set to 1 then
   pixel i j of img is set to transp. *)
let set_transparency img mask l c =
  let lmask = Array.length mask in
  if lmask <> 0 then
  let cmask = Array.length mask.(0) in
  for i = 0 to l - 1 do
    if i < lmask then
      let line = img.(i) in
      let mask_line = mask.(i) in
      for j = 0 to c - 1 do
        if j < cmask then
        if mask_line.(j) = white then line.(j) <- transp
      done
  done
;;

let rec read_ppm_ib ib =
  let mn, l, c = read_ppm_header ib in
  let img =
    match mn with
    | P1 -> read_ascii_pbm_ib ib l c
    | P4 -> read_raw_pbm_ib ib l c
    | (P2 | P3 | P5 | P6) ->
       let _max = read_max ib in
       match mn with
       | P2 -> read_ascii_pgm_ib ib l c
       | P3 -> read_ascii_ppm_ib ib l c
       | P5 -> read_raw_pgm_ib ib l c
       | P6 -> read_raw_ppm_ib ib l c
       | P4 | P1 -> assert false in
  try
    if Scanf.Scanning.end_of_input ib then img else
    let mask = read_ppm_ib ib in
    set_transparency img mask l c;
    img
  with
  | _ -> img
;;

let read_ppm fname =
  let ib = Scanf.Scanning.from_file_bin fname in
  try read_ppm_ib ib with
  | End_of_file -> invalid_arg "read_ppm: premature end of file"
;;

let load_pixmap = read_ppm;;
let load_ppm = read_ppm;;
let load_bitmap = read_ppm;;

(* Saving images. *)

let save_ppm_header mn oc l c =
  output_string oc (Printf.sprintf "%s\n" (string_of_magic_number mn));
  output_string oc "# CREATOR: Caml graphic package\n";
  output_string oc (Printf.sprintf "%d %d\n" c l);
  if mn <> P1 && mn <> P4 then output_string oc (Printf.sprintf "%d\n" 255)
;;

let bit_set = 1 and bit_cleared = 0;;

let gen_save_raw_pbm_oc is_white img oc l c =
  save_ppm_header P4 oc l c;
  for i = 0 to l - 1 do
    let line = img.(i) in
    let rec loop j bn byte =
      if j = c then
       if bn = 0 then () else
         let byte = byte lsl (8 - bn) in
         output_byte oc byte else
       if bn = 8 then (output_byte oc byte; loop j 0 0) else
      let color =
        if is_white line.(j) then bit_set else bit_cleared in
      let new_byte = (byte lsl 1) lor color in
      loop (j + 1) (bn + 1) new_byte in
    loop 0 0 0
 done
;;

(* Save a bitmap in raw form. *)
let save_raw_pbm_oc = gen_save_raw_pbm_oc (fun c -> c = white);;

(* Save a transparency mask as a bitmap in raw form. *)
let save_mask = gen_save_raw_pbm_oc (fun c -> c = transp);;

(* Save a pixmap in raw form. *)
let save_raw_ppm_oc img oc l c =
  save_ppm_header P6 oc l c;
  let has_transp = ref false in
  for i = 0 to l - 1 do
    let line = img.(i) in
    for j = 0 to c - 1 do
      let color = line.(j) in
      if color = transp then has_transp := true;
      let b = color land 0xFF in
      let g = (color lsr 8) land 0xFF in
      let r = (color lsr 16) land 0xFF in
      output_byte oc r;
      output_byte oc g;
      output_byte oc b
    done
  done;
  !has_transp
;;

let save_ppm_oc img oc =
  let l = Array.length img in
    if l = 0 then invalid_arg "save_ppm: invalid null line number";
  let c = Array.length img.(0) in
  if c = 0 then invalid_arg "save_ppm: invalid null column number";
  let has_transp = save_raw_ppm_oc img oc l c in
  if has_transp then
    begin
     (* Save the transparency mask *)
     output_char oc '\n';
     save_mask img oc l c;
    end
;;

let save_ppm fname img =
  let oc = open_out_bin fname in
  save_ppm_oc img oc;
  close_out oc
;;

let save_pixmap = save_ppm;;

let save_bitmap_oc img oc =
  let l = Array.length img in
    if l = 0 then invalid_arg "save_ppm: invalid null line number";
  let c = Array.length img.(0) in
  if c = 0 then invalid_arg "save_ppm: invalid null column number";
  save_raw_pbm_oc img oc l c
;;

let save_bitmap fname img =
  let oc = open_out_bin fname in
  save_bitmap_oc img oc;
  close_out oc
;;

let load_generic_image fname =
  Pixmap.generic_image_of_pixmap (load_pixmap fname)
;;

let save_generic_image fname img =
  save_pixmap fname (Pixmap.pixmap_of_generic_image img)
;;

let load_image fname =
  let ppm = load_ppm fname in
  Graphics.make_image ppm
;;

let save_image fname img =
  let pixels = Graphics.dump_image img in
  save_ppm fname pixels
;;

let ppm_of_generic_image = Pixmap.pixmap_of_generic_image;;
let generic_image_of_ppm = Pixmap.generic_image_of_pixmap;;

let image_of_ppm = Pixmap.image_of_pixmap;;
let ppm_of_image = Pixmap.pixmap_of_image;;
