(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2013-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: worker.ml,v 1.7 2014-03-14 15:47:26 fclement Exp $ *)

(* Generic part of workers for the inversion platform. *)

module Nat = Unatural;;

module Cmd_opt = Ucommand_option;;

(* Printing facilities. *)

type 'a printer = 'a Format_printer.printer;;

type 'a string_printer = string -> 'a -> string;;
type ('a, 'b) string_printer_lifter = 'a printer -> 'b string_printer;;

let format_sprint print ident x =
  let ppf = Format.str_formatter in
  Format.fprintf ppf "@[<hv 4>@ %s =@ %a@]" ident print x;
  Format.flush_str_formatter ()
;;

let sprint_nat = format_sprint Unatural_print.print_t;;
let sprint_float = format_sprint Lib_print.print_quoted_float;;

let sprint_homogeneous_couple print =
  format_sprint (Lib_print.print_Couple print print)
;;
let sprint_array print = format_sprint (Lib_print.print_Array print);;
let sprint_array_array print = sprint_array (Lib_print.print_Array print);;
let sprint_matrix print = format_sprint (Umatrix_print.print_t print);;

let sprint_nat_couple = sprint_homogeneous_couple Unatural_print.print_t;;
let sprint_nat_matrix = sprint_matrix Unatural_print.print_t;;

let sprint_float_array_array = sprint_array_array Lib_print.print_quoted_float;;
let sprint_float_matrix = sprint_matrix Lib_print.print_quoted_float;;

let um_qual = Upervasives.qualify "User_model";;

let um_invalid_arg = Upervasives.invalid_arg_qual um_qual;;
let um_failwith = Upervasives.failwith_qual um_qual;;

(* Tasks. *)

let coarse_parameter_to_fine_parameter ppm maa =
  Umatrix.map_ij (fun i _j pp_ij -> maa.(i).(Unatural.to_int pp_ij)) ppm
;;

let check_same_dimensions fun_name msg dim dim' =
  if dim' <> dim then um_invalid_arg fun_name msg
;;

let check_valid_dimension fun_name dim dim' =
  let msg = Printf.sprintf "dimension must be %d" (Unatural.to_int dim) in
  check_same_dimensions fun_name msg dim dim'
;;

let check_valid_zone_numbers fun_name ppm maa =
  let valid_zone_numbers =
    Uarray_ng.array_for_all_i
      (fun i ->
        Uarray.for_all
          (fun pp_ij -> pp_ij < Uarray.length (Uarray_ng.array_get maa i)))
      (Umatrix.to_array_array ppm) in
  if not valid_zone_numbers then
    um_invalid_arg fun_name "zone numbers in partition are not all valid"
;;

let make_init
    init_engine optim ({ Udebug.spf_debug = debug; _; } as spf) dim =
  debug (sprint_nat "dim" dim);
  let (sz, ppm1, maa1_init) = init_engine spf dim in
  debug (sprint_nat_couple "sz" sz);
  debug (sprint_nat_matrix "ppm1" ppm1);
  let (maa1, j1) = optim spf (ppm1, maa1_init) in
  debug (sprint_float_array_array "maa1" maa1);
  debug (sprint_float "j1" j1);
  (sz, ppm1, maa1, j1)
;;

let make_grad grad_engine ({ Udebug.spf_debug = debug; _; } as spf) pm =
  debug (sprint_float_matrix "pm" pm);
  let gm = grad_engine spf pm in
  debug (sprint_float_matrix "gm" gm);
  gm
;;

let make_optim
    optim_engine ({ Udebug.spf_debug = debug; _; } as spf) (ppm, maa_init) =
  debug (sprint_nat_matrix "ppm" ppm);
  debug (sprint_float_array_array "maa_init" maa_init);
  let dim = Uarray.length maa_init in
  let dim' = Umatrix.dim_i ppm in
  let msg =
    "arguments partition and initial coarse parameter \
     have incompatible dimensions" in
  check_same_dimensions "optim" msg dim dim';
  check_valid_zone_numbers "optim" ppm maa_init;
  let ((maa_opt, j_opt) as res) = optim_engine spf ppm maa_init in
  debug (sprint_float_array_array "maa_opt" maa_opt);
  debug (sprint_float "j_opt" j_opt);
  res
;;

let make_final
    final_engine ({ Udebug.spf_debug = debug; _; } as spf) pm =
  debug (sprint_float_matrix "pm" pm);
  final_engine spf pm;
;;

(* Wrappers. *)

let mk_init_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._1_ then
    um_failwith "init_proj" "initialization should receive one value" else
  report "one value received";
  let dim = Nat.of_int (Typed_value.to_int vals.(0)) in
  report "dimension extracted";
  dim
;;

let mk_init_inj _spf ((ni, nj), ppm1, maa1, j1) =
  let sz = (Nat.to_int ni, Nat.to_int nj) in
  let ppm1 = Umatrix.map Nat.to_int ppm1 in
  let (lens1, a1) = Uarray_ng.array_array_pack maa1 in
  let lens1 = Array.map Nat.to_int lens1 in
  [| Typed_value.of_int_couple sz;
     Typed_value.of_c_int_matrix ppm1;
     Typed_value.of_int_vector lens1;
     Typed_value.of_float_vector a1;
     Typed_value.of_float j1; |]
;;

let make_init_wrapper =
  Pio.make_task_wrapper "init" mk_init_inj mk_init_proj
;;

let mk_grad_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._1_ then
    um_failwith "grad_proj" "gradient should receive one value" else
  report "one value received";
  let pm = Typed_value.to_c_float_matrix vals.(0) in
  report "fine parameter extracted";
  pm
;;

let mk_grad_inj _spf gm = [| Typed_value.of_c_float_matrix gm; |];;

let make_grad_wrapper =
  Pio.make_task_wrapper "grad" mk_grad_inj mk_grad_proj
;;

let mk_optim_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._3_ then
    um_failwith "optim_proj" "optimization should receive three values" else
  report "three values received";
  let ppm = Typed_value.to_c_int_matrix vals.(0) in
  let ppm = Umatrix.map Nat.of_int ppm in
  report "partition extracted";
  let lens_init = Typed_value.to_int_vector vals.(1) in
  let lens_init = Array.map Nat.of_int lens_init in
  report "coarse parameter lengths extracted";
  let a_init = Typed_value.to_float_vector vals.(2) in
  let maa_init = Uarray_ng.array_array_unpack lens_init a_init in
  report "initial coarse parameter extracted";
  (ppm, maa_init)
;;

let mk_optim_inj _spf (maa_opt, j_opt) =
  let (lens_opt, a_opt) = Uarray_ng.array_array_pack maa_opt in
  let lens_opt = Array.map Nat.to_int lens_opt in
  [| Typed_value.of_int_vector lens_opt;
     Typed_value.of_float_vector a_opt;
     Typed_value.of_float j_opt; |]
;;

let make_optim_wrapper =
  Pio.make_task_wrapper "optim" mk_optim_inj mk_optim_proj
;;

let mk_final_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._1_ then
    um_failwith "final_proj" "finalization should receive one value" else
  report "one value received";
  let pm = Typed_value.to_c_float_matrix vals.(0) in
  report "fine parameter extracted";
  pm
;;

let mk_final_inj _spf _ = [| Typed_value.of_string "\nOK, bye!\n"; |];;

let make_final_wrapper =
  Pio.make_task_wrapper "final" mk_final_inj mk_final_proj
;;

(* Task computation functions. *)

let make_initialize init pf piob =
  Pio.worker_wrap_task piob (make_init_wrapper pf "init") init
;;

let make_compute_gradient grad pf piob =
  Pio.worker_wrap_task piob (make_grad_wrapper pf "grad") grad
;;

let make_optimize optim pf piob =
  Pio.worker_wrap_task piob (make_optim_wrapper pf "optim") optim
;;

let make_finalize final pf piob vals =
  Pio.worker_wrap_task piob (make_final_wrapper pf "final") final vals;
  (raise Exit : unit);
;;

(* Main loop. *)

let mk_task_tbl init grad optim final =
  Ulist1.of_list
    [
     ("init", make_initialize init);
     ("grad", make_compute_gradient grad);
     ("optim", make_optimize optim);
     ("final", make_finalize final);
    ]
;;

let main_loop welcome_action farewell_action init grad optim final =
  let task_tbl = mk_task_tbl init grad optim final in
  Pio.worker_main welcome_action farewell_action task_tbl
;;
